/* Copyright (C) 2006-2017 Michel Gutierrez <mig@downloadhelper.net>
 * All Rights Reserved
 * 
 * If you are interested in methods used in this project, follow the
 * open source project at https://github.com/mi-g/fang
 */

var videos = {};
var image_re = new RegExp("url\\(\"(.*)\"\\)");
var extract_vid_re = new RegExp("/videos/([0-9]+)");
var contentHeight = 0;

function GetVideoData(videoId,url) {
	videos[url] = 1;
	var xmlhttp = new XMLHttpRequest();
	xmlhttp.onreadystatechange = function() {
		if (xmlhttp.readyState == XMLHttpRequest.DONE && xmlhttp.status == 200) {
			var resp = JSON.parse(xmlhttp.responseText);
			var message = {
				fromContent: true,
				type: 'tbsn-video-data',
				url: url,
			}
			if(resp.source) message.source = resp.source;
			if(resp.name) message.name = resp.name;
			if(resp.picture) message.picture = resp.picture;
			if(resp.from && resp.from.name) message.fromName = resp.from.name;
			if(resp.length) message.duration = Math.round(resp.length);
			window.postMessage(message,'*');
		}
	}
	xmlhttp.open("GET", "https://graph.facebook.com/"+videoId+"?fields=name,picture,source,from,length", true);
	xmlhttp.send();
}

function Search(elem,url) {
	var node = elem.parentNode;
	while(node) {
		var vlink = node.querySelector("a[data-video-id]");
		if(vlink) {
			var videoId = vlink.getAttribute("data-video-id");
			GetVideoData(videoId,url);
			return;
		}
		var plinks = node.querySelectorAll("a[class='profileLink']");
		for(var i=0;i<plinks.length;i++) {
			var plink = plinks[i];
			var href = plink.getAttribute("href");
			if(href) {
				var m = extract_vid_re.exec(href);
				if(m) {
					var videoId = m[1];
					GetVideoData(videoId,url);
					return;
				}
			}
		}
		node = node.parentNode;
	}
	return null;
}

function GetThumbnail(elem,url) {
	var img = elem.nextSibling && elem.nextSibling.firstChild;
	if(!img || img.nodeName!="IMG")
		return;
	var m = image_re.exec(img.style.backgroundImage);
	if(m)
		window.postMessage({
			fromContent: true,
			type: 'tbsn-video-data',
			url: url,
			picture: m[1],
		},'*');
}

function CheckVideos() {
	if(document.body.clientHeight<=contentHeight)
		return;
	contentHeight=document.body.clientHeight;
	var elems = document.querySelectorAll("video");
	for(var i=0; i<elems.length; i++) {
		var elem = elems[i];
		var src = elem.getAttribute("src");
		if(src && !videos[src]) {
			GetThumbnail(elem,src);
			Search(elem,src);
		}
	}
}

CheckVideos();

setInterval(CheckVideos,1000);
