/* Copyright (C) 2006-2017 Michel Gutierrez <mig@downloadhelper.net>
 * All Rights Reserved
 * 
 * If you are interested in methods used in this project, follow the
 * open source project at https://github.com/mi-g/fang
 */

const simplePrefs = require('sdk/simple-prefs');
const simpleStorage = require('sdk/simple-storage');
const prefsService = require('sdk/preferences/service');
const self = require('sdk/self');
const timers = require("sdk/timers");
const prefName = "extensions.dwhelper";

var syncLegacyDataPort = null;
var storageUpdateTimers = {};

exports.syncStorage = function(store) {
	if(syncLegacyDataPort) {
		if(storageUpdateTimers[store])
			timers.clearTimeout(storageUpdateTimers[store]);
		storageUpdateTimers[store] = timers.setTimeout(()=>{
			delete storageUpdateTimers[store];
			syncLegacyDataPort.postMessage({
				["v6store-"+store]: simpleStorage.storage[store]
			});	
		},100);
	} else
		console.warn("syncStorage(",store,") called but port not set");
}

exports.setSyncLegacyDataPort = function(port) {

	syncLegacyDataPort = port;

	var message = {};

	prefsService.keys(prefName).forEach(function(pref) {
		pref = pref.substring(prefName.length+1);
		message["v6pref-"+pref] = simplePrefs.prefs[pref];
	});
	Object.keys(simpleStorage.storage).forEach((store)=>{
		message["v6store-"+store] = simpleStorage.storage[store];
	});
	port.postMessage(message);

	simplePrefs.on("", (pref) => {
		port.postMessage({
			["v6pref-"+pref]: simplePrefs.prefs[pref]
		});
	});

}


